#!/usr/bin/perl
#
# script used to analyze the output of the logging function
# in atemu

use warnings;
use Class::Struct;

struct FuncCall => {
    addr     => '$',
    sh       => '$',
    ptsindex => '$',
};

my @maxStackSize;             # max stack size for each chip
my $stacklinecount = 0;       # total number of stack lines in file
my $linecount = 0;            # total number of lines in file
my @intrNestCnt;              # interrupt depth for each chip
my @callStack;                # call stack for the grpah chip
my $tasklinecount = 0;        # number of TASK lines in file
my $intrlinecount = 0;        # number of INTR lines in file
my $calllinecount = 0;        # number of CALL lines in file
my $startcycle = 0;           # start cycle to graph from, supplied as arg
my $endcycle = 0;             # last cycle to graph
my $laststackcycle = 0;       # cycle number of last stack line in file we graphed
my $lastintrcycle = 0;        # ""        "          INTR   "       "       "
my $laststacksize;            # ie current stack size, last stack height we saw
my @lastTaskCycle;            # last time we saw a task cycle for each chip; used to measure task len
my @lastTaskPC;               #  " " "                  PC
my @lastTaskSH;               # " "         "       "   stack height
my $showtaskinfo = 0;         # option flags to show XXX info
my $showintrinfo = 0;
my $showcallinfo = 0;
my $showsimintrinfo = 0;
my %functions = ();           # map of function addresses to names, from nm
my @pts;                      # array of the jgraph polygon points, each one is a function call instance
my $currpts = 0;
my $graphing = 0;             # true if we have plotted some points and are still in cycle range
my $tempfn = "temp22932.dump";

# Graph Constants
my $cyclespan = 1000;
my $xaxis = 9.5;                 #inches
my $yaxis = 4.0;

# Argument Processing
if ($ARGV[0] =~ /help/) {
    print "Stack Height Grapher for atemu log files\n";
    print "Writes graph output to encapsulated postscript file <filename>.ps and jgraph file (.jgr)\n";
    print "Requires log files to be logged with -SC, and JGraph\n";
    print "usage:\tstackgraph [-OPTIONS] [chip] [startcycle] [inputfile]\n";
    print "\tchip\t\tChip number to graph\n";
    print "\tstartcycle\tCycle to start graphing from\n";
    print "\tinputfile\tatemu log file\n\n";
    print "Options:\n\t-t\t\tDisplay TinyOS Task info\n";
    print "\t-i\t\tDisplay interrupt nesting info\n";
    print "\t-c <filename>\tLookup function call symbol info in executable specified\n";
    print "\t-s\t\tDisplay info about simultaneous interrupts only\n";
    exit(0);
}

$i = 0;
if ($ARGV[0] =~ /^-/) {
    if ($ARGV[0] =~ /t/) {
	$showtaskinfo = 1;
    }
    if ($ARGV[0] =~ /i/) {
	$showintrinfo = 1;
    }
    if ($ARGV[0] =~ /c/) {
	$showcallinfo = 1;
	$execfile = $ARGV[1];
	$i++;
    }
    if ($ARGV[0] =~ /s/) {
	$showsimintrinfo = 1;
    }
    $i++;
}

$number = $ARGV[$i];
if ($number =~ /^[1234567890]/) {
    $graphchip = $number;
}
else {
    print "Please input the chip number to graph.\nUse 'stackgraph --help' for usage info\n";
    exit;
}

$i++;
$number = $ARGV[$i];
if ($number =~ /^[1234567890]/) {
    $startcycle = $number;
    $endcycle = $number + $cyclespan;
    $laststackcycle = $startcycle;
    $lastintrcycle = $startcycle;
    print "startcycle: $startcycle endcycle: $endcycle\n";
}
else {
    print "Please input a valid cycle number to start graphing from.\nUse 'stackgraph --help' for usage info\n";
    exit;
}

my $file;
if ($i != $#ARGV) {
    $file = $ARGV[$#ARGV];
    die unless open(LOGFILE, "bzcat $file |");
} else {
    print "Please specify a input log file.\nUse 'stackgraph --help' for usage info\n";
    exit;
}

@temp = split /\./, $file;
$jgrfile = $temp[0];
$jgrfile .= ".jgr";
die "Can't open jgraph output file.\n" unless open(JGRFILE, ">$jgrfile");

# get symbol information
if ($showcallinfo) {
    system("nm -n $execfile > $tempfn");
    die "Can't open symbol output file; probably didn't execute nm correctly or insufficient write privleges\n" 
	unless open(SYMBOL, "$tempfn");
    
    while (<SYMBOL>) {
	@words = split;
	$functions{hex($words[0])} = $words[2];
	#print "mapped $words[0] ($numb) to $words[2]\n";
    }
    close SYMBOL;
}

# JGraph Header stuff
print JGRFILE "(* stackgraph.pl output *)\n\n";
print JGRFILE "newgraph\n";
print JGRFILE "xaxis size $xaxis min $startcycle max $endcycle hash 100 mhash 4 label : Cycles\n";
print JGRFILE "yaxis size $yaxis min 0 max 160 hash 10 mhash 2 label : Stack Height\n";

#dummy init call
my $tempcall = FuncCall->new();
$tempcall->addr(0);
$tempcall->sh(0);
$tempcall->ptsindex(-1);
push(@callStack, $tempcall);

while (<LOGFILE>) {
    
    @words = split;
    $linecount++;

    $chipno = $words[0];
    $cycles = $words[2];
    if ($words[1] =~ /STAK/) {
	$stacklinecount++;
	$stacksize =  $words[$#words];
	if ($chipno > $#maxStackSize) {
	    $maxStackSize[$chipno] = 0;
	    $intrNest[$chipno] = 0;
	    $lastTaskCycle[$chipno] = 0;
	    $lastTaskPC[$chipno] = 0;
	    $lastTaskSH[$chipno] = 0;
	}
	else {
	    # 255 is ruled out here as this is the initial stack size as the chip boots up
	    if ($maxStackSize[$chipno] < $stacksize && $stacksize != 255) {
		$maxStackSize[$chipno] = $stacksize;
		#print "max stack size: $maxStackSize[$chipno]\n";
	    }
	}

	if ($chipno == $graphchip) {
	    $laststacksize = $stacksize;

	    # detect start of graphing and create polygons with a clean start edge
	    if ($cycles >= $startcycle && $cycles <= $endcycle && $graphing == 0) {
		for ($i = 0; $i <= $#callStack; $i++) {
		    if ($callStack[$i]->ptsindex == -1) {  #add point info for funct's not listed already
			$#pts++;
			$callStack[$i]->ptsindex($#pts);
			if ($i < $#callStack) {
			    $above = ${callStack[$i + 1]}->sh;
			} else {
			    $above = $stacksize;
			}
		  
			if (defined $functions{${callStack[$i]}->addr} ) {
			    $name = $functions{${callStack[$i]}->addr};
			} elsif (${callStack[$i]}->addr < 256) {
			    $addr = ${callStack[$i]}->addr;
			    $name = "INT $addr";
			} else {
			    $addr = ${callStack[$i]}->addr;
			    $name = "0x$addr";
			}
			$tempsh = ${callStack[$i]}->sh;
			$pts[$#pts] = "label : $name\npts $startcycle $tempsh $startcycle $above ";
		    }
		}
		$graphing = 1;
	    }

	    # detect end of graphing and add a clean edge to the existing polygons
	    if ($cycles > $endcycle && $graphing == 1) {
		for ($i = $#callStack; $i >= 0; $i--) {
		    if ($i == $#callStack) {
			$above = $stacksize;
		    } else {
			$above = ${callStack[$i + 1]}->sh;
		    }
		    $tempsh = ${callStack[$i]}->sh;
		    $pts[$callStack[$i]->ptsindex] .= "$endcycle $above $endcycle $tempsh ";
		}
		$graphing = 0;
	    }
	}

	# End of TASK detection
	$pc = hex($words[3]);
	if ($pc - $lastTaskPC[$chipno] <= 6 && $pc - $lastTaskPC[$chipno] >= 0) {
	    $gap = $cycles - $lastTaskCycle[$chipno];
	    if ($showtaskinfo) {
		print "Chip $chipno END TASK - cycles during this task: $gap\n";
	    }
	    $lastTaskCycle[$chipno] = $cycles;
	    $lastTaskPC[$chipno] = 0;
	    $lastTaskSH[$chipno] = 0;
	}

	# add a point to the current curve if we're graphing
	if ($cycles >= $startcycle && $cycles <= $endcycle && $chipno == $graphchip) {
	    @temp = split / /, $pts[$currpts];
	    $lastpnt = $temp[$#temp];
	    $pts[$currpts] .= "$cycles $lastpnt $cycles $stacksize ";
	}
	
    }
    elsif ($words[1] =~ /INTR/) {
	$intrlinecount++;
	if ($showintrinfo) {
	    for ($i = 0; $i < $intrNestCnt[$chipno]; $i++) {
		print "\t";
	    }
	    print "Chip $chipno INTR $words[4] cycle $cycles\n";
	}
#	if ($words[4] != 18) {
#	    print "INT $words[4] fired at $cycles\n";
#	}
	if ($chipno == $graphchip) {
	    if ($cycles >= $startcycle && $cycles <= $endcycle) {
		$#pts++;
		$currpts = $#pts;
		$pts[$currpts] = "label : INT $words[4]\npts $cycles $laststacksize ";
	    }

	    my $tempcall = FuncCall->new();
	    # in intrStack addr is just the interrupt number
	    $tempcall->addr($words[4]);
	    $tempcall->sh($laststacksize);
	    $tempcall->ptsindex($#pts);

	    push(@callStack, $tempcall);
	}
	$intrNestCnt[$chipno]++;
	if ($showsimintrinfo && $intrNestCnt[$chipno] > 1) {
	    print "Simulatenous intterrupts on chip $chipno at $cycles, INT $words[4] fired\n";
	}
    }
    elsif ($words[1] =~ /RETI/) {
	$intrNestCnt[$chipno]--;
	if ($chipno == $graphchip) {
	    $tempcall = pop @callStack;
	    if ($graphing) {
		if ($#callStack > -1) {       # add a point to the end to smooth graph polygons
		    $prevstacksize = $tempcall->sh;
		    $pts[$currpts] .= "$cycles $prevstacksize ";
		} else {
		    $pts[$currpts] .= "$cycles 0 ";
		}
	    }

	    $currpts = $callStack[$#callStack]->ptsindex;
	}
	if ($showintrinfo) {
	    for ($i = 0; $i < $intrNestCnt[$chipno]; $i++) {
		print "\t";
	    }
	    print "Chip $chipno RETI cycle $cycles\n";
	}
    }
    elsif ($words[1] =~ /CALL/ || $words[1] =~ /TASK/) {
	$calllinecount++;
	if ($chipno == $graphchip) {
	    if ($cycles >= $startcycle && $cycles <= $endcycle) {
		$#pts++;
		$currpts = $#pts;
		$addr = hex($words[4]) * 2;
		if (defined $functions{$addr}) {
		    $name = $functions{$addr};
		} else {
		    $name = "0x$addr";
		}
		$pts[$currpts] = "label : $name\npts $cycles $laststacksize ";
	    }

	    my $tempcall = FuncCall->new();
	    $tempcall->addr(hex($words[4]) * 2);
	    $tempcall->sh($laststacksize);
	    $tempcall->ptsindex($#pts);

	    push(@callStack, $tempcall);
	}
	if ($words[1] =~ /TASK/) {       # assembly: ICALL
	    $tasklinecount++;
	    $gap = $cycles - $lastTaskCycle[$chipno];
	    if ($showtaskinfo) {
		print "Chip $chipno TASK - cycles between tasks: $gap, cycles $cycles\n";
	    }
	    $lastTaskCycle[$chipno] = $cycles;
	    $lastTaskPC[$chipno] = hex($words[3]);
	    if ($chipno == $graphchip) {
		$lastTaskSH[$chipno] = $laststacksize;
	    }
	}  
    }    
    elsif ($words[1] =~ /RETC/) {         # assembly: RET
	if ($chipno == $graphchip) {
	    $tempcall = pop @callStack;
	    if ($graphing) {
		if ($#callStack > -1) {       # add a point to the end to smooth graph polygons
		    $prevstacksize = $tempcall->sh;
		    $pts[$currpts] .= "$cycles $prevstacksize ";
		} else {
		    $pts[$currpts] .= "$cycles 0 ";		
		}
	    }

	    $currpts = $callStack[$#callStack]->ptsindex;
	}
    }
  
}

print "Total number of lines: $linecount\n";
print "Total number of stack lines: $stacklinecount\n";
print "Total number of interrupt lines: $intrlinecount\n";
print "Total number of task lines: $tasklinecount\n";
for ($i = 0; $i <= $#maxStackSize; $i++) {
    print "Chip $i\n------------------------\nhighest stack size: $maxStackSize[$i]\n";
    print "net interrupt nesting: $intrNest[$i]\n";
}

close LOGFILE;

for ($j = 0; $j <= $#pts; $j++) {
    print JGRFILE "newline poly linethickness 2 marksize 1 ";

    @colors = ( "0 0 1", "0 .5 .5", "0 1 0", ".5 0 0",
                ".5 .5 1", ".25 .25 .25", "0 .5 0", "1 1 0", ".75 0 .75", "1 0 0",
                ".5 .5 .5", ".5 1 0", ".5 .5 0", "0 .25 .5", "1 .6 1", ".8 .5 .2",  "1 .5 .5");

    $color = $colors[$j % $#colors];
    if ($pts[$j] =~ /INT/) {
	print JGRFILE "pcfill $color color $color ppattern estripe ";
    } else {
	print JGRFILE "pcfill $color color $color ";
    }

    print JGRFILE "$pts[$j]\n";
} 

print JGRFILE "\ntitle : Chip $graphchip - $file\n";
close JGRFILE;

@temp = split /\./, $file;
$outfilename = $temp[0];
$outfilename .= ".eps";
system("jgraph $jgrfile > $outfilename\n");
system("rm -f $tempfn\n");


